# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

from aniDBmaper import AniDBMaper


class ResponseResolver(object):
    def __init__(self, data):
        restag, rescode, resstr, datalines = self.parse(data)

        self.restag = restag
        self.rescode = rescode
        self.resstr = resstr
        self.datalines = datalines

    def parse(self, data):
        resline = data.split('\n', 1)[0]
        lines = data.split('\n')[1:-1]

        rescode, resstr = resline.split(' ', 1)
        if rescode[0] == 'T':
            restag = rescode
            rescode, resstr = resstr.split(' ', 1)
        else:
            restag = None

        datalines = []
        for line in lines:
            datalines.append(line.split('|'))

        return restag, rescode, resstr, datalines

    def resolve(self, cmd):
        return responses[self.rescode](cmd, self.restag, self.rescode, self.resstr, self.datalines)


class Response:
    def __init__(self, cmd, restag, rescode, resstr, rawlines):
        self.req = cmd
        self.restag = restag
        self.rescode = rescode
        self.resstr = resstr
        self.rawlines = rawlines
        self.maper = AniDBMaper()
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()

    def __repr__(self):
        tmp = "%s(%s,%s,%s) %s\n" % (
            self.__class__.__name__, repr(self.restag), repr(self.rescode), repr(self.resstr), repr(self.attrs))

        m = 0
        for line in self.datalines:
            for k, v in line.items():
                if len(k) > m:
                    m = len(k)

        for line in self.datalines:
            tmp += "  Line:\n"
            for k, v in line.items():
                tmp += "    %s:%s %s\n" % (k, (m - len(k)) * ' ', v)
        return tmp

    def parse(self):
        tmp = self.resstr.split(' ', len(self.codehead))
        self.attrs = dict(zip(self.codehead, tmp[:-1]))
        self.resstr = tmp[-1]

        self.datalines = []
        for rawline in self.rawlines:
            normal = dict(zip(self.codetail, rawline))
            rawline = rawline[len(self.codetail):]
            rep = []
            if len(self.coderep):
                while rawline:
                    tmp = dict(zip(self.coderep, rawline))
                    rawline = rawline[len(self.coderep):]
                    rep.append(tmp)
            # normal['rep']=rep
            self.datalines.append(normal)

    def handle(self):
        if self.req:
            self.req.handle(self)


class LoginAcceptedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        sesskey	- session key
        address	- your address (ip:port) as seen by the server

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'LOGIN_ACCEPTED'
        self.codetail = ()
        self.coderep = ()

        nat = cmd.parameters['nat']
        nat = int(nat is None and nat or '0')
        if nat:
            self.codehead = ('sesskey', 'address')
        else:
            self.codehead = ('sesskey',)


class LoginAcceptedNewVerResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        sesskey	- session key
        address	- your address (ip:port) as seen by the server

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'LOGIN_ACCEPTED_NEW_VER'
        self.codetail = ()
        self.coderep = ()

        nat = cmd.parameters['nat']
        nat = int(nat is None and nat or '0')
        if nat:
            self.codehead = ('sesskey', 'address')
        else:
            self.codehead = ('sesskey',)


class LoggedOutResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'LOGGED_OUT'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class ResourceResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'RESOURCE'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class StatsResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'STATS'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class TopResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'TOP'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class UptimeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        uptime	- udpserver uptime in milliseconds

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'UPTIME'
        self.codehead = ()
        self.codetail = ('uptime',)
        self.coderep = ()


class EncryptionEnabledResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        salt	- salt

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'ENCRYPTION_ENABLED'
        self.codehead = ('salt',)
        self.codetail = ()
        self.coderep = ()


class MylistEntryAddedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        entrycnt - number of entries added

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'MYLIST_ENTRY_ADDED'
        self.codehead = ()
        self.codetail = ('entrycnt',)
        self.coderep = ()


class MylistEntryDeletedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        entrycnt - number of entries

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'MYLIST_ENTRY_DELETED'
        self.codehead = ()
        self.codetail = ('entrycnt',)
        self.coderep = ()


class AddedFileResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'ADDED_FILE'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class AddedStreamResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'ADDED_STREAM'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class EncodingChangedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'ENCODING_CHANGED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class FileResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        eid		episode id
        gid		group id
        lid		mylist id
        state		state
        size		size
        ed2k		ed2k
        md5		md5
        sha1		sha1
        crc32		crc32
        dublang		dub language
        sublang		sub language
        quality		quality
        source		source
        audiocodec	audio codec
        audiobitrate	audio bitrate
        videocodec	video codec
        videobitrate	video bitrate
        resolution	video resolution
        filetype	file type (extension)
        length		length in seconds
        description	description
        filename	anidb file name
        gname		group name
        gshortname	group short name
        epno		number of episode
        epname		ep english name
        epromaji	ep romaji name
        epkanji		ep kanji name
        totaleps	anime total episodes
        lastep		last episode nr (highest, not special)
        year		year
        type		type
        romaji		romaji name
        kanji		kanji name
        name		english name
        othername	other name
        shortnames	short name list
        synonyms	synonym list
        categories	category list
        relatedaids	related aid list
        producernames	producer name list
        producerids	producer id list

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'FILE'
        self.codehead = ()
        self.coderep = ()

        fmask = cmd.parameters['fmask']
        amask = cmd.parameters['amask']

        codeListF = self.maper.getFileCodesF(fmask)
        codeListA = self.maper.getFileCodesA(amask)
        # print "File - codelistF: "+str(codeListF)
        # print "File - codelistA: "+str(codeListA)


        self.codetail = tuple(['fid'] + codeListF + codeListA)


class MylistResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        lid	 - mylist id
        fid	 - file id
        eid	 - episode id
        aid	 - anime id
        gid	 - group id
        date	 - date when you added this to mylist
        state	 - the location of the file
        viewdate - date when you marked this watched
        storage	 - for example the title of the cd you have this on
        source	 - where you got the file (bittorrent,dc++,ed2k,...)
        other	 - other data regarding this file
        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'MYLIST'
        self.codehead = ()
        self.codetail = ('lid', 'fid', 'eid', 'aid', 'gid', 'date', 'state', 'viewdate', 'storage', 'source', 'other')
        self.coderep = ()


class MylistStatsResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        animes		- animes
        eps		- eps
        files		- files
        filesizes	- size of files
        animesadded	- added animes
        epsadded	- added eps
        filesadded	- added files
        groupsadded	- added groups
        leechperc	- leech %
        lameperc	- lame %
        viewedofdb	- viewed % of db
        mylistofdb	- mylist % of db
        viewedofmylist	- viewed % of mylist
        viewedeps	- number of viewed eps
        votes		- votes
        reviews		- reviews

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'MYLIST_STATS'
        self.codehead = ()
        self.codetail = (
            'animes', 'eps', 'files', 'filesizes', 'animesadded', 'epsadded', 'filesadded', 'groupsadded', 'leechperc',
            'lameperc', 'viewedofdb', 'mylistofdb', 'viewedofmylist', 'viewedeps', 'votes', 'reviews')
        self.coderep = ()


class NotificationResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        nid         - notification id
        unknown     - unsure what this parameter is
        type        - Type of notification: type=>  0=all, 1=new, 2=group, 3=complete
        aid         - anime id
        priority    - low = 0, medium = 1, high = 2 (unconfirmed)
        date        - date notification subscribed to

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFICATION'
        self.codehead = ()
        self.codetail = ('nid', 'unknown', 'type', 'aid', 'priority', 'date')
        self.coderep = ()


class AnimeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'ANIME'
        self.codehead = ()
        self.coderep = ()

        # TODO: impl random anime
        amask = cmd.parameters['amask']
        codeList = self.maper.getAnimeCodesA(amask)
        self.codetail = tuple(codeList)


class AnimeBestMatchResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'ANIME_BEST_MATCH'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class RandomanimeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'RANDOMANIME'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class EpisodeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        eid	- episode id
        aid	- anime id
        length	- length
        rating	- rating
        votes	- votes
        epno	- number of episode
        name	- english name of episode
        romaji	- romaji name of episode
        kanji	- kanji name of episode

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'EPISODE'
        self.codehead = ()
        self.codetail = ('eid', 'aid', 'length', 'rating', 'votes', 'epno', 'name', 'romaji', 'kanji')
        self.coderep = ()


class ProducerResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        pid	  - producer id
        name	  - name of producer
        shortname - short name
        othername - other name
        type	  - type
        pic	  - picture name
        url	  - home page url

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'PRODUCER'
        self.codehead = ()
        self.codetail = ('pid', 'name', 'shortname', 'othername', 'type', 'pic', 'url')
        self.coderep = ()


class GroupResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        gid	   - group id
        rating	   - rating
        votes	   - votes
        animes	   - anime count
        files	   - file count
        name	   - name
        shortname  - short
        ircchannel - irc channel
        ircserver  - irc server
        url	   - url

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'GROUP'
        self.codehead = ()
        self.codetail = (
            'gid', 'rating', 'votes', 'animes', 'files', 'name', 'shortname', 'ircchannel', 'ircserver', 'url')
        self.coderep = ()


class GroupstatusResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        gid	   - group id
        rating	   - rating
        votes	   - votes
        animes	   - anime count
        files	   - file count
        name	   - name
        shortname  - short
        ircchannel - irc channel
        ircserver  - irc server
        url	   - url

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'GROUPSTATUS'
        self.codehead = ()
        self.codetail = ('gid', 'name', 'state', ' last_episode_number', 'rating', 'votes', 'episode_range')
        self.coderep = ()


class BuddyListResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        start	- mylist entry number of first buddy on this packet
        end	- mylist entry number of last buddy on this packet
        total	- total number of buddies on mylist

        data:
        uid	- uid
        name	- username
        state	- state

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'BUDDY_LIST'
        self.codehead = ('start', 'end', 'total')
        self.codetail = ('uid', 'username', 'state')
        self.coderep = ()


class BuddyStateResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        start	- mylist entry number of first buddy on this packet
        end	- mylist entry number of last buddy on this packet
        total	- total number of buddies on mylist

        data:
        uid	- uid
        state	- online state

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'BUDDY_STATE'
        self.codehead = ('start', 'end', 'total')
        self.codetail = ('uid', 'state')
        self.coderep = ()


class BuddyAddedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'BUDDY_ADDED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BuddyDeletedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'BUDDY_DELETED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BuddyAcceptedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'BUDDY_ACCEPTED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BuddyDeniedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'BUDDY_DENIED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class VotedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        name	- aname/ename/gname

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'VOTED'
        self.codehead = ()
        self.codetail = ('name',)
        self.coderep = ()


class VoteFoundResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        name	- aname/ename/gname
        value	- vote value

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'VOTE_FOUND'
        self.codehead = ()
        self.codetail = ('name', 'value')
        self.coderep = ()


class VoteUpdatedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        name	- aname/ename/gname
        value	- vote value

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'VOTE_UPDATED'
        self.codehead = ()
        self.codetail = ('name', 'value')
        self.coderep = ()


class VoteRevokedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        name	- aname/ename/gname
        value	- vote value

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'VOTE_REVOKED'
        self.codehead = ()
        self.codetail = ('name', 'value')
        self.coderep = ()


class NotificationAddedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        nid - notofication id

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFICATION_ITEM_ADDED'
        self.codehead = ()
        self.codetail = ('nid')
        self.coderep = ()


class NotificationDeletedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        nid - notofication id

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFICATION_ITEM_DELETED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NotificationUpdatedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        nid - notofication id

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFICATION_ITEM_UPDATED'
        self.codehead = ()
        self.codetail = ('nid')
        self.coderep = ()


class MultipleNotificationResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        nid         - notification id
        type        -
        priority    -


        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'MULTIPLE NOTIFICATION ITEMS'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ('nid', 'type', 'priority',)


class NotificationEnabledResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFICATION_ENABLED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NotificationNotifyResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        nid	- notify packet id

        data:
        aid	- anime id
        date	- date
        count	- count
        name	- name of the anime

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFICATION_NOTIFY'
        self.codehead = ('nid',)
        self.codetail = ('aid', 'date', 'count', 'name')
        self.coderep = ()


class NotificationMessageResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        nid	- notify packet id

        data:
        type	- type
        date	- date
        uid	- user id of the sender
        name	- name of the sender
        subject	- subject

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFICATION_MESSAGE'
        self.codehead = ('nid',)
        self.codetail = ('type', 'date', 'uid', 'name', 'subject')
        self.coderep = ()


class NotificationBuddyResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        nid	- notify packet id

        data:
        uid	- buddy uid
        type	- event type

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFICATION_BUDDY'
        self.codehead = ('notify_packet_id',)
        self.codetail = ('uid', 'type')
        self.coderep = ()


class NotificationShutdownResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:
        nid	- notify packet id

        data:
        time	- time offline
        comment	- comment

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFICATION_SHUTDOWN'
        self.codehead = ('nid',)
        self.codetail = ('time', 'comment')
        self.coderep = ()


class PushackConfirmedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'PUSHACK_CONFIRMED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NotifyackSuccessfulMResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFYACK_SUCCESSFUL_M'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NotifyackSuccessfulNResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFYACK_SUCCESSFUL_N'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NotificationStateResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        notifies - pending notifies
        msgs	 - pending msgs
        buddys	 - number of online buddys

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFICATION'
        self.codehead = ()
        self.coderep = ()

        buddy = cmd.parameters['buddy']
        buddy = int(buddy is not None and buddy or '0')
        if buddy:
            self.codetail = ('notifies', 'msgs', 'buddys')
        else:
            self.codetail = ('notifies', 'msgs')


class NotifylistResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        type	- type
        nid	- notify id

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFYLIST'
        self.codehead = ()
        self.codetail = ('type', 'nid')
        self.coderep = ()


class NotifygetMessageResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        nid	- notify id
        uid	- from user id
        uname	- from username
        date	- date
        type	- type
        title	- title
        body	- body

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFYGET_MESSAGE'
        self.codehead = ()
        self.codetail = ('nid', 'uid', 'uname', 'date', 'type', 'title', 'body')
        self.coderep = ()


class NotifygetNotifyResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        aid	- aid
        type	- type
        count	- count
        date	- date
        name	- anime name
        fid     - comma separated list of fids
        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFYGET_NOTIFY'
        self.codehead = ()
        self.codetail = ('aid', 'type', 'count', 'date', 'name', 'fid')
        self.coderep = ()


class SendmsgSuccessfulResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'SENDMSG_SUCCESSFUL'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class UserResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        uid	- user id

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'USER'
        self.codehead = ()
        self.codetail = ('uid',)
        self.coderep = ()


class PongResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'PONG'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class AuthpongResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'AUTHPONG'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchResourceResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_RESOURCE'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class ApiPasswordNotDefinedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'API_PASSWORD_NOT_DEFINED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class FileAlreadyInMylistResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'FILE_ALREADY_IN_MYLIST'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class MylistEntryEditedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        entries	- number of entries edited

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'MYLIST_ENTRY_EDITED'
        self.codehead = ()
        self.codetail = ('entries',)
        self.coderep = ()


class MultipleMylistEntriesResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        name	   - anime title
        eps	   - episodes
        unknowneps - eps with state unknown
        hddeps	   - eps with state on hdd
        cdeps	   - eps with state on cd
        deletedeps - eps with state deleted
        watchedeps - watched eps
        gshortname - group short name
        geps	   - eps for group

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'MULTIPLE_MYLIST_ENTRIES'
        self.codehead = ()
        self.codetail = ('name', 'eps', 'unknowneps', 'hddeps', 'cdeps', 'deletedeps', 'watchedeps')
        self.coderep = ('gshortname', 'geps')


class SizeHashExistsResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'SIZE_HASH_EXISTS'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class InvalidDataResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'INVALID_DATA'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class StreamnoidUsedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'STREAMNOID_USED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchFileResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_FILE'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchEntryResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_ENTRY'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class MultipleFilesFoundResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        fid	- file id

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'MULTIPLE_FILES_FOUND'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ('fid',)


class NoSuchNotificationResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_NOTIFICATION'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoGroupsFoundResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO GROUPS FOUND'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchAnimeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_ANIME'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchEpisodeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_EPISODE'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchProducerResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_PRODUCER'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchGroupResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_GROUP'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BuddyAlreadyAddedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'BUDDY_ALREADY_ADDED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchBuddyResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_BUDDY'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BuddyAlreadyAcceptedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'BUDDY_ALREADY_ACCEPTED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BuddyAlreadyDeniedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'BUDDY_ALREADY_DENIED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchVoteResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_VOTE'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class InvalidVoteTypeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'INVALID_VOTE_TYPE'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class InvalidVoteValueResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'INVALID_VOTE_VALUE'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class PermvoteNotAllowedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        aname	- name of the anime

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'PERMVOTE_NOT_ALLOWED'
        self.codehead = ()
        self.codetail = ('aname',)
        self.coderep = ()


class AlreadyPermvotedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        name	- aname/ename/gname

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'ALREADY_PERMVOTED'
        self.codehead = ()
        self.codetail = ('name',)
        self.coderep = ()


class NotificationDisabledResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOTIFICATION_DISABLED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchPacketPendingResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_PACKET_PENDING'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchEntryMResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_ENTRY_M'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchEntryNResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_ENTRY_N'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchMessageResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_MESSAGE'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchNotifyResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_NOTIFY'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchUserResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_USER'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoChanges(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_CHANGES'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NotLoggedInResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NOT_LOGGED_IN'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchMylistFileResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_MYLIST_FILE'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchMylistEntryResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_MYLIST_ENTRY'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class LoginFailedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'LOGIN_FAILED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class LoginFirstResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'LOGIN_FIRST'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class AccessDeniedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'ACCESS_DENIED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class ClientVersionOutdatedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'CLIENT_VERSION_OUTDATED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class ClientBannedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'CLIENT_BANNED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class IllegalInputOrAccessDeniedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'ILLEGAL_INPUT_OR_ACCESS_DENIED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class InvalidSessionResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'INVALID_SESSION'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class NoSuchEncryptionTypeResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'NO_SUCH_ENCRYPTION_TYPE'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class EncodingNotSupportedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'ENCODING_NOT_SUPPORTED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class BannedResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'BANNED'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class UnknownCommandResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'UNKNOWN_COMMAND'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class InternalServerErrorResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'INTERNAL_SERVER_ERROR'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class AnidbOutOfServiceResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'ANIDB_OUT_OF_SERVICE'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class ServerBusyResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'SERVER_BUSY'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class ApiViolationResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'API_VIOLATION'
        self.codehead = ()
        self.codetail = ()
        self.coderep = ()


class VersionResponse(Response):
    def __init__(self, cmd, restag, rescode, resstr, datalines):
        """
        attributes:

        data:
        version	- server version

        """
        Response.__init__(self, cmd, restag, rescode, resstr, datalines)
        self.codestr = 'VERSION'
        self.codehead = ()
        self.codetail = ('version',)
        self.coderep = ()


responses = {
    '200': LoginAcceptedResponse,
    '201': LoginAcceptedNewVerResponse,
    '203': LoggedOutResponse,
    '205': ResourceResponse,
    '206': StatsResponse,
    '207': TopResponse,
    '208': UptimeResponse,
    '209': EncryptionEnabledResponse,
    '210': MylistEntryAddedResponse,
    '211': MylistEntryDeletedResponse,
    '214': AddedFileResponse,
    '215': AddedStreamResponse,
    '219': EncodingChangedResponse,
    '220': FileResponse,
    '221': MylistResponse,
    '222': MylistStatsResponse,
    '224': NotificationResponse,
    '225': GroupstatusResponse,
    '230': AnimeResponse,
    '231': AnimeBestMatchResponse,
    '232': RandomanimeResponse,
    '240': EpisodeResponse,
    '245': ProducerResponse,
    '246': NotificationAddedResponse,
    '247': NotificationDeletedResponse,
    '248': NotificationUpdatedResponse,
    '249': MultipleNotificationResponse,
    '250': GroupResponse,
    '253': BuddyListResponse,
    '254': BuddyStateResponse,
    '255': BuddyAddedResponse,
    '256': BuddyDeletedResponse,
    '257': BuddyAcceptedResponse,
    '258': BuddyDeniedResponse,
    '260': VotedResponse,
    '261': VoteFoundResponse,
    '262': VoteUpdatedResponse,
    '263': VoteRevokedResponse,
    '270': NotificationEnabledResponse,
    '271': NotificationNotifyResponse,
    '272': NotificationMessageResponse,
    '273': NotificationBuddyResponse,
    '274': NotificationShutdownResponse,
    '280': PushackConfirmedResponse,
    '281': NotifyackSuccessfulMResponse,
    '282': NotifyackSuccessfulNResponse,
    '290': NotificationStateResponse,
    '291': NotifylistResponse,
    '292': NotifygetMessageResponse,
    '293': NotifygetNotifyResponse,
    '294': SendmsgSuccessfulResponse,
    '295': UserResponse,
    '300': PongResponse,
    '301': AuthpongResponse,
    '305': NoSuchResourceResponse,
    '309': ApiPasswordNotDefinedResponse,
    '310': FileAlreadyInMylistResponse,
    '311': MylistEntryEditedResponse,
    '312': MultipleMylistEntriesResponse,
    '314': SizeHashExistsResponse,
    '315': InvalidDataResponse,
    '316': StreamnoidUsedResponse,
    '320': NoSuchFileResponse,
    '321': NoSuchEntryResponse,
    '322': MultipleFilesFoundResponse,
    '324': NoSuchNotificationResponse,
    '325': NoGroupsFoundResponse,
    '330': NoSuchAnimeResponse,
    '340': NoSuchEpisodeResponse,
    '345': NoSuchProducerResponse,
    '350': NoSuchGroupResponse,
    '355': BuddyAlreadyAddedResponse,
    '356': NoSuchBuddyResponse,
    '357': BuddyAlreadyAcceptedResponse,
    '358': BuddyAlreadyDeniedResponse,
    '360': NoSuchVoteResponse,
    '361': InvalidVoteTypeResponse,
    '362': InvalidVoteValueResponse,
    '363': PermvoteNotAllowedResponse,
    '364': AlreadyPermvotedResponse,
    '370': NotificationDisabledResponse,
    '380': NoSuchPacketPendingResponse,
    '381': NoSuchEntryMResponse,
    '382': NoSuchEntryNResponse,
    '392': NoSuchMessageResponse,
    '393': NoSuchNotifyResponse,
    '394': NoSuchUserResponse,
    '399': NoChanges,
    '403': NotLoggedInResponse,
    '410': NoSuchMylistFileResponse,
    '411': NoSuchMylistEntryResponse,
    '500': LoginFailedResponse,
    '501': LoginFirstResponse,
    '502': AccessDeniedResponse,
    '503': ClientVersionOutdatedResponse,
    '504': ClientBannedResponse,
    '505': IllegalInputOrAccessDeniedResponse,
    '506': InvalidSessionResponse,
    '509': NoSuchEncryptionTypeResponse,
    '519': EncodingNotSupportedResponse,
    '555': BannedResponse,
    '598': UnknownCommandResponse,
    '600': InternalServerErrorResponse,
    '601': AnidbOutOfServiceResponse,
    '602': ServerBusyResponse,
    '666': ApiViolationResponse,
    '998': VersionResponse
}



# -*- coding: latin-1 -*-
# Author: adaur <adaur.underground@gmail.com>
# URL: https://sickrage.ca
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import cookielib
import re
import urllib

import requests

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import bs4_parser
from sickrage.providers import TorrentProvider


class XthorProvider(TorrentProvider):
    def __init__(self):

        super(XthorProvider, self).__init__("Xthor", "http://xthor.bz", True)

        self.supports_backlog = True

        self.cj = cookielib.CookieJar()

        self.urls.update({
            'search': "{base_url}/browse.php?search=%s%s".format(base_url=self.urls['base_url'])
        })

        self.categories = "&searchin=title&incldead=0"

        self.username = None
        self.password = None
        self.ratio = None

        self.cache = TVCache(self, min_time=10)

    def login(self):
        if any(requests.utils.dict_from_cookiejar(sickrage.srCore.srWebSession.cookies).values()):
            return True

        login_params = {'username': self.username,
                        'password': self.password,
                        'submitme': 'X'}

        try:
            response = sickrage.srCore.srWebSession.post(self.urls['base_url'] + '/takelogin.php', data=login_params,
                                                         timeout=30).text
        except Exception:
            sickrage.srCore.srLogger.warning("[{}]: Unable to connect to provider".format(self.name))
            return False

        if not re.search('donate.php', response):
            sickrage.srCore.srLogger.warning(
                "[{}]: Invalid username or password. Check your settings".format(self.name))
            return False

        return True

    def search(self, search_params, search_mode='eponly', epcount=0, age=0, epObj=None):

        results = []
        items = {'Season': [], 'Episode': [], 'RSS': []}

        # check for auth
        if not self.login():
            return results

        for mode in search_params.keys():
            sickrage.srCore.srLogger.debug("Search Mode: %s" % mode)
            for search_string in search_params[mode]:

                if mode != 'RSS':
                    sickrage.srCore.srLogger.debug("Search string: %s " % search_string)

                searchURL = self.urls['search'] % (urllib.quote(search_string), self.categories)
                sickrage.srCore.srLogger.debug("Search URL: %s" % searchURL)

                try:
                    data = sickrage.srCore.srWebSession.get(searchURL, cache=False).text
                except Exception:
                    sickrage.srCore.srLogger.debug("No data returned from provider")
                    continue

                with bs4_parser(data) as html:
                    resultsTable = html.find("table", {"class": "table2 table-bordered2"})
                    if resultsTable:
                        rows = resultsTable.findAll("tr")
                        for row in rows:
                            link = row.find("a", href=re.compile("details.php"))
                            if link:
                                title = link.text
                                download_url = self.urls['base_url'] + '/' + \
                                               row.find("a", href=re.compile("download.php"))['href']
                                # FIXME
                                size = -1
                                seeders = 1
                                leechers = 0

                                if not all([title, download_url]):
                                    continue

                                # Filter unseeded torrent
                                # if seeders < self.minseed or leechers < self.minleech:
                                #    if mode != 'RSS':
                                #        LOGGER.debug(u"Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(title, seeders, leechers))
                                #    continue

                                item = title, download_url, size, seeders, leechers
                                if mode != 'RSS':
                                    sickrage.srCore.srLogger.debug("Found result: %s " % title)

                                items[mode].append(item)

            # For each search mode sort all the items by seeders if available if available
            items[mode].sort(key=lambda tup: tup[3], reverse=True)

            results += items[mode]

        return results

    def seed_ratio(self):
        return self.ratio



import glob
import io
import os
import shutil

from setuptools import setup, Command

# Get the version number
with io.open(os.path.abspath(os.path.join(os.path.dirname(__file__), 'sickrage', 'version.txt'))) as f:
    version = f.read()


def requires():
    with open(os.path.abspath(os.path.join(os.path.dirname(__file__), 'requirements.txt'))) as f:
        return f.read().splitlines()


class CleanCommand(Command):
    """Custom clean command to tidy up the project root."""
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        shutil.rmtree(os.path.abspath(os.path.join(os.path.dirname(__file__), 'build')), ignore_errors=True)
        shutil.rmtree(os.path.abspath(os.path.join(os.path.dirname(__file__), '*.pyc')), ignore_errors=True)
        shutil.rmtree(os.path.abspath(os.path.join(os.path.dirname(__file__), '*.tgz')), ignore_errors=True)
        shutil.rmtree(os.path.abspath(os.path.join(os.path.dirname(__file__), 'sickrage.egg-info')), ignore_errors=True)
        [os.remove(f) for f in glob.glob("dist/sickrage-*")]

setup(
    name='sickrage',
    version=version,
    description='Automatic Video Library Manager for TV Shows',
    author='echel0n',
    author_email='echel0n@sickrage.ca',
    license='GPLv3',
    url='https://git.sickrage.ca',
    keywords=['sickrage', 'sickragetv', 'tv', 'torrent', 'nzb', 'video', 'echel0n'],
    packages=['sickrage'],
    install_requires=requires(),
    include_package_data=True,
    platforms='any',
    zip_safe=False,
    test_suite='tests',
    cmdclass={
        'clean': CleanCommand
    },
    entry_points={
        "console_scripts": [
            "sickrage=sickrage:main"
        ]
    }
)



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import os
import shutil
import stat

import UnRAR2
from UnRAR2.rar_exceptions import ArchiveHeaderBroken, FileOpenError, \
    IncorrectRARPassword, InvalidRARArchive, InvalidRARArchiveUsage

import sickrage
from sickrage.core.common import Quality
from sickrage.core.exceptions import EpisodePostProcessingFailedException, \
    FailedPostProcessingFailedException
from sickrage.core.helpers import isMediaFile, isRarFile, isSyncFile, \
    is_hidden_folder, notTorNZBFile, real_path
from sickrage.core.nameparser import InvalidNameException, InvalidShowException, \
    NameParser
from sickrage.core.processors import failed_processor, post_processor


class ProcessResult(object):
    def __init__(self):
        self.result = True
        self.output = ''
        self.missedfiles = []
        self.aggresult = True


def delete_folder(folder, check_empty=True):
    """
    Removes a folder from the filesystem

    :param folder: Path to folder to remove
    :param check_empty: Boolean, check if the folder is empty before removing it, defaults to True
    :return: True on success, False on failure
    """

    # check if it's a folder
    if not os.path.isdir(folder):
        return False

    # check if it isn't TV_DOWNLOAD_DIR
    if sickrage.srCore.srConfig.TV_DOWNLOAD_DIR:
        if real_path(folder) == real_path(sickrage.srCore.srConfig.TV_DOWNLOAD_DIR):
            return False

    # check if it's empty folder when wanted checked
    try:
        if check_empty:
            check_files = os.listdir(folder)
            if check_files:
                sickrage.srCore.srLogger.info(
                    "Not deleting folder {} found the following files: {}".format(folder, check_files))
                return False

            sickrage.srCore.srLogger.info("Deleting folder (if it's empty): " + folder)
            shutil.rmtree(folder)
        else:
            sickrage.srCore.srLogger.info("Deleting folder: " + folder)
            shutil.rmtree(folder)
    except (OSError, IOError) as e:
        sickrage.srCore.srLogger.warning("Warning: unable to delete folder: {}: {}".format(folder, e))
        return False

    return True


def delete_files(processPath, notwantedFiles, result, force=False):
    """
    Remove files from filesystem

    :param processPath: path to process
    :param notwantedFiles: files we do not want
    :param result: Processor results
    :param force: Boolean, force deletion, defaults to false
    """

    if not result.result and force:
        result.output += logHelper("Forcing deletion of files, even though last result was not success",
                                   sickrage.srCore.srLogger.DEBUG)
    elif not result.result:
        return

    # Delete all file not needed
    for cur_file in notwantedFiles:
        cur_file_path = os.path.join(processPath, cur_file)

        if not os.path.isfile(cur_file_path):
            continue  # Prevent error when a notwantedfiles is an associated files

        result.output += logHelper("Deleting file %s" % cur_file, sickrage.srCore.srLogger.DEBUG)

        # check first the read-only attribute
        file_attribute = os.stat(cur_file_path)[0]
        if not file_attribute & stat.S_IWRITE:
            # File is read-only, so make it writeable
            result.output += logHelper("Changing ReadOnly Flag for file %s" % cur_file, sickrage.srCore.srLogger.DEBUG)
            try:
                os.chmod(cur_file_path, stat.S_IWRITE)
            except OSError as e:
                result.output += logHelper(
                    "Cannot change permissions of %s: %s" % (
                        cur_file, str(e.strerror).decode(sickrage.srCore.SYS_ENCODING)),
                    sickrage.srCore.srLogger.DEBUG)
        try:
            os.remove(cur_file_path)
        except OSError as e:
            result.output += logHelper(
                "Unable to delete file %s: %s" % (cur_file, str(e.strerror).decode(sickrage.srCore.SYS_ENCODING)),
                sickrage.srCore.srLogger.DEBUG)


def logHelper(logMessage, logLevel=None):
    sickrage.srCore.srLogger.log(logLevel or sickrage.srCore.srLogger.INFO, logMessage)
    return logMessage + "\n"


def processDir(dirName, nzbName=None, process_method=None, force=False, is_priority=None, delete_on=False, failed=False,
               proc_type="auto", **kwargs):
    """
    Scans through the files in dirName and processes whatever media files it finds

    :param dirName: The folder name to look in
    :param nzbName: The NZB name which resulted in this folder being downloaded
    :param force: True to postprocess already postprocessed files
    :param failed: Boolean for whether or not the download failed
    :param proc_type: Type of postprocessing auto or manual
    """

    result = ProcessResult()

    result.output += logHelper("Processing folder %s" % dirName, sickrage.srCore.srLogger.DEBUG)

    result.output += logHelper("TV_DOWNLOAD_DIR: %s" % sickrage.srCore.srConfig.TV_DOWNLOAD_DIR,
                               sickrage.srCore.srLogger.DEBUG)
    postpone = False

    # if they passed us a real dir then assume it's the one we want
    if os.path.isdir(dirName):
        dirName = os.path.realpath(dirName)

    # if the client and SickRage are not on the same machine translate the Dir in a network dir
    elif sickrage.srCore.srConfig.TV_DOWNLOAD_DIR and os.path.isdir(sickrage.srCore.srConfig.TV_DOWNLOAD_DIR) \
            and os.path.normpath(dirName) != os.path.normpath(sickrage.srCore.srConfig.TV_DOWNLOAD_DIR):
        dirName = os.path.join(sickrage.srCore.srConfig.TV_DOWNLOAD_DIR,
                               os.path.abspath(dirName).split(os.path.sep)[-1])
        result.output += logHelper("Trying to use folder %s" % dirName, sickrage.srCore.srLogger.DEBUG)

    # if we didn't find a real dir then quit
    if not os.path.isdir(dirName):
        result.output += logHelper(
            "Unable to figure out what folder to process. If your downloader and SiCKRAGE aren't on the same PC make sure you fill out your TV download dir in the config.",
            sickrage.srCore.srLogger.DEBUG)
        return result.output

    path, dirs, files = get_path_dir_files(dirName, nzbName, proc_type)

    files = [x for x in files if notTorNZBFile(x)]
    SyncFiles = [x for x in files if isSyncFile(x)]

    # Don't post process if files are still being synced and option is activated
    if SyncFiles and sickrage.srCore.srConfig.POSTPONE_IF_SYNC_FILES:
        postpone = True

    nzbNameOriginal = nzbName

    if not postpone:
        result.output += logHelper("PostProcessing Path: %s" % path, sickrage.srCore.srLogger.INFO)
        result.output += logHelper("PostProcessing Dirs: [%s]" % ", ".join(dirs), sickrage.srCore.srLogger.DEBUG)

        rarFiles = [x for x in files if isRarFile(x)]
        rarContent = unRAR(path, rarFiles, force, result)
        files += rarContent
        videoFiles = [x for x in files if isMediaFile(x)]
        videoInRar = [x for x in rarContent if isMediaFile(x)]

        result.output += logHelper("PostProcessing Files: [%s]" % ", ".join(files), sickrage.srCore.srLogger.DEBUG)
        result.output += logHelper("PostProcessing VideoFiles: [%s]" % ", ".join(videoFiles),
                                   sickrage.srCore.srLogger.DEBUG)
        result.output += logHelper("PostProcessing RarContent: [%s]" % ", ".join(rarContent),
                                   sickrage.srCore.srLogger.DEBUG)
        result.output += logHelper("PostProcessing VideoInRar: [%s]" % ", ".join(videoInRar),
                                   sickrage.srCore.srLogger.DEBUG)

        # If nzbName is set and there's more than one videofile in the folder, files will be lost (overwritten).
        if len(videoFiles) >= 2:
            nzbName = None

        if not process_method:
            process_method = sickrage.srCore.srConfig.PROCESS_METHOD

        result.result = True

        # Don't Link media when the media is extracted from a rar in the same path
        if process_method in ('hardlink', 'symlink') and videoInRar:
            process_media(path, videoInRar, nzbName, 'move', force, is_priority, result)
            delete_files(path, rarContent, result)
            for video in set(videoFiles) - set(videoInRar):
                process_media(path, [video], nzbName, process_method, force, is_priority, result)
        elif sickrage.srCore.srConfig.DELRARCONTENTS and videoInRar:
            process_media(path, videoInRar, nzbName, process_method, force, is_priority, result)
            delete_files(path, rarContent, result, True)
            for video in set(videoFiles) - set(videoInRar):
                process_media(path, [video], nzbName, process_method, force, is_priority, result)
        else:
            for video in videoFiles:
                process_media(path, [video], nzbName, process_method, force, is_priority, result)

    else:
        result.output += logHelper("Found temporary sync files, skipping post processing for: %s" % path)
        result.output += logHelper("Sync Files: [%s] in path %s" % (", ".join(SyncFiles), path))
        result.missedfiles.append("%s : Syncfiles found" % path)

    # Process Video File in all TV Subdir
    for curDir in [x for x in dirs if validateDir(path, x, nzbNameOriginal, failed, result)]:

        result.result = True

        for processPath, _, fileList in os.walk(os.path.join(path, curDir), topdown=False):

            if not validateDir(path, processPath, nzbNameOriginal, failed, result):
                continue

            postpone = False

            SyncFiles = [x for x in fileList if isSyncFile(x)]

            # Don't post process if files are still being synced and option is activated
            if SyncFiles and sickrage.srCore.srConfig.POSTPONE_IF_SYNC_FILES:
                postpone = True

            if not postpone:
                rarFiles = [x for x in fileList if isRarFile(x)]
                rarContent = unRAR(processPath, rarFiles, force, result)
                fileList = set(fileList + rarContent)
                videoFiles = [x for x in fileList if isMediaFile(x)]
                videoInRar = [x for x in rarContent if isMediaFile(x)]
                notwantedFiles = [x for x in fileList if x not in videoFiles]
                if notwantedFiles:
                    result.output += logHelper("Found unwanted files: [%s]" % ", ".join(notwantedFiles),
                                               sickrage.srCore.srLogger.DEBUG)

                # Don't Link media when the media is extracted from a rar in the same path
                if process_method in ('hardlink', 'symlink') and videoInRar:
                    process_media(processPath, videoInRar, nzbName, 'move', force, is_priority, result)
                    process_media(processPath, set(videoFiles) - set(videoInRar), nzbName, process_method, force,
                                  is_priority, result)
                    delete_files(processPath, rarContent, result)
                elif sickrage.srCore.srConfig.DELRARCONTENTS and videoInRar:
                    process_media(processPath, videoInRar, nzbName, process_method, force, is_priority, result)
                    process_media(processPath, set(videoFiles) - set(videoInRar), nzbName, process_method, force,
                                  is_priority, result)
                    delete_files(processPath, rarContent, result, True)
                else:
                    process_media(processPath, videoFiles, nzbName, process_method, force, is_priority, result)

                    # Delete all file not needed
                    if process_method != "move" or not result.result \
                            or (
                                            proc_type == "manual" and not delete_on):  # Avoid to delete files if is Manual PostProcessing
                        continue

                    delete_files(processPath, notwantedFiles, result)

                    if (
                                not sickrage.srCore.srConfig.NO_DELETE or proc_type == "manual") and process_method == "move" and \
                                    os.path.normpath(processPath) != os.path.normpath(
                                sickrage.srCore.srConfig.TV_DOWNLOAD_DIR):
                        if delete_folder(processPath, check_empty=True):
                            result.output += logHelper("Deleted folder: %s" % processPath,
                                                       sickrage.srCore.srLogger.DEBUG)
            else:
                result.output += logHelper("Found temporary sync files, skipping post processing for: %s" % processPath)
                result.output += logHelper("Sync Files: [%s] in path %s" % (", ".join(SyncFiles), processPath))
                result.missedfiles.append("%s : Syncfiles found" % processPath)

    if result.aggresult:
        result.output += logHelper("Processing completed")
        if result.missedfiles:
            result.output += logHelper("I did encounter some unprocessable items: [%s]" % ", ".join(result.missedfiles))
    else:
        result.output += logHelper(
            "Problem(s) during processing, failed the following files/folders:  [%s]" % ", ".join(
                result.missedfiles),
            sickrage.srCore.srLogger.WARNING)

    return result.output


def validateDir(path, dirName, nzbNameOriginal, failed, result):
    """
    Check if directory is valid for processing

    :param path: Path to use
    :param dirName: Directory to check
    :param nzbNameOriginal: Original NZB name
    :param failed: Previously failed objects
    :param result: Previous results
    :return: True if dir is valid for processing, False if not
    """

    IGNORED_FOLDERS = ['.AppleDouble', '.@__thumb', '@eaDir']

    folder_name = os.path.basename(dirName)
    if folder_name in IGNORED_FOLDERS:
        return False

    result.output += logHelper("Processing folder " + dirName, sickrage.srCore.srLogger.DEBUG)

    if folder_name.startswith('_FAILED_'):
        result.output += logHelper("The directory name indicates it failed to extract.", sickrage.srCore.srLogger.DEBUG)
        failed = True
    elif folder_name.startswith('_UNDERSIZED_'):
        result.output += logHelper("The directory name indicates that it was previously rejected for being undersized.",
                                   sickrage.srCore.srLogger.DEBUG)
        failed = True
    elif folder_name.upper().startswith('_UNPACK'):
        result.output += logHelper(
            "The directory name indicates that this release is in the process of being unpacked.",
            sickrage.srCore.srLogger.DEBUG)
        result.missedfiles.append(dirName + " : Being unpacked")
        return False

    if failed:
        process_failed(os.path.join(path, dirName), nzbNameOriginal, result)
        result.missedfiles.append(dirName + " : Failed download")
        return False

    if is_hidden_folder(os.path.join(path, dirName)):
        result.output += logHelper("Ignoring hidden folder: " + dirName, sickrage.srCore.srLogger.DEBUG)
        result.missedfiles.append(dirName + " : Hidden folder")
        return False

    # make sure the dir isn't inside a show dir
    for dbData in [x['doc'] for x in sickrage.srCore.mainDB.db.all('tv_shows', with_doc=True)]:
        if dirName.lower().startswith(os.path.realpath(dbData["location"]).lower() + os.sep) or \
                        dirName.lower() == os.path.realpath(dbData["location"]).lower():
            result.output += logHelper(
                "Cannot process an episode that's already been moved to its show dir, skipping " + dirName,
                sickrage.srCore.srLogger.WARNING)
            return False

    # Get the videofile list for the next checks
    allFiles = []
    allDirs = []
    for _, processdir, fileList in os.walk(os.path.join(path, dirName), topdown=False):
        allDirs += processdir
        allFiles += fileList

    videoFiles = [x for x in allFiles if isMediaFile(x)]
    allDirs.append(dirName)

    # check if the dir have at least one tv video file
    for video in videoFiles:
        try:
            NameParser().parse(video, cache_result=False)
            return True
        except (InvalidNameException, InvalidShowException):
            pass

    for proc_dir in allDirs:
        try:
            NameParser().parse(proc_dir, cache_result=False)
            return True
        except (InvalidNameException, InvalidShowException):
            pass

    if sickrage.srCore.srConfig.UNPACK:
        # Search for packed release
        packedFiles = [x for x in allFiles if isRarFile(x)]

        for packed in packedFiles:
            try:
                NameParser().parse(packed, cache_result=False)
                return True
            except (InvalidNameException, InvalidShowException):
                pass

    result.output += logHelper(dirName + " : No processable items found in folder", sickrage.srCore.srLogger.DEBUG)
    return False


def unRAR(path, rarFiles, force, result):
    """
    Extracts RAR files

    :param path: Path to look for files in
    :param rarFiles: Names of RAR files
    :param force: process currently processing items
    :param result: Previous results
    :return: List of unpacked file names
    """

    unpacked_files = []

    if sickrage.srCore.srConfig.UNPACK and rarFiles:

        result.output += logHelper("Packed Releases detected: " + str(rarFiles), sickrage.srCore.srLogger.DEBUG)

        for archive in rarFiles:

            result.output += logHelper("Unpacking archive: " + archive, sickrage.srCore.srLogger.DEBUG)

            try:
                rar_handle = UnRAR2.RarFile(os.path.join(path, archive))

                # Skip extraction if any file in archive has previously been extracted
                skip_file = False
                for file_in_archive in [os.path.basename(x.filename) for x in rar_handle.infolist() if not x.isdir]:
                    if already_postprocessed(path, file_in_archive, force, result):
                        result.output += logHelper(
                            "Archive file already post-processed, extraction skipped: " + file_in_archive,
                            sickrage.srCore.srLogger.DEBUG)
                        skip_file = True
                        break

                if skip_file:
                    continue

                rar_handle.extract(path=path, withSubpath=False, overwrite=False)
                for x in rar_handle.infolist():
                    if not x.isdir:
                        basename = os.path.basename(x.filename)
                        if basename not in unpacked_files:
                            unpacked_files.append(basename)
                del rar_handle

            except ArchiveHeaderBroken as e:
                result.output += logHelper("Failed Unrar archive {0}: Unrar: Archive Header Broken".format(archive),
                                           sickrage.srCore.srLogger.ERROR)
                result.result = False
                result.missedfiles.append(archive + " : Unpacking failed because the Archive Header is Broken")
                continue
            except IncorrectRARPassword:
                result.output += logHelper("Failed Unrar archive {0}: Unrar: Incorrect Rar Password".format(archive),
                                           sickrage.srCore.srLogger.ERROR)
                result.result = False
                result.missedfiles.append(archive + " : Unpacking failed because of an Incorrect Rar Password")
                continue
            except FileOpenError:
                result.output += logHelper(
                    "Failed Unrar archive {0}: Unrar: File Open Error, check the parent folder and destination file permissions.".format(
                        archive), sickrage.srCore.srLogger.ERROR)
                result.result = False
                result.missedfiles.append(archive + " : Unpacking failed with a File Open Error (file permissions?)")
                continue
            except InvalidRARArchiveUsage:
                result.output += logHelper("Failed Unrar archive {0}: Unrar: Invalid Rar Archive Usage".format(archive),
                                           sickrage.srCore.srLogger.ERROR)
                result.result = False
                result.missedfiles.append(archive + " : Unpacking Failed with Invalid Rar Archive Usage")
                continue
            except InvalidRARArchive:
                result.output += logHelper("Failed Unrar archive {0}: Unrar: Invalid Rar Archive".format(archive),
                                           sickrage.srCore.srLogger.ERROR)
                result.result = False
                result.missedfiles.append(archive + " : Unpacking Failed with an Invalid Rar Archive Error")
                continue
            except Exception as e:
                result.output += logHelper("Failed Unrar archive {}: {}".format(archive, e),
                                           sickrage.srCore.srLogger.ERROR)
                result.result = False
                result.missedfiles.append(archive + " : Unpacking failed for an unknown reason")
                continue

        result.output += logHelper("UnRar content: " + str(unpacked_files), sickrage.srCore.srLogger.DEBUG)

    return unpacked_files


def already_postprocessed(dirName, videofile, force, result):
    """
    Check if we already post processed a file

    :param dirName: Directory a file resides in
    :param videofile: File name
    :param force: Force checking when already checking (currently unused)
    :param result: True if file is already postprocessed, False if not
    :return:
    """
    if force:
        return False

    # Avoid processing the same dir again if we use a process method <> move
    if [x for x in sickrage.srCore.mainDB.db.all('tv_episodes', with_doc=True)
        if x['doc']['release_name'] == dirName]:
        return True
    else:
        if [x for x in sickrage.srCore.mainDB.db.all('tv_episodes', with_doc=True)
            if x['doc']['release_name'] == [videofile.rpartition('.')[0]]]: return True

        # Needed if we have downloaded the same episode @ different quality
        # But we need to make sure we check the history of the episode we're going to PP, and not others
        np = NameParser(dirName, tryIndexers=True)
        try:
            parse_result = np.parse(dirName)
        except:
            parse_result = False

        for h in [h['doc'] for h in sickrage.srCore.mainDB.db.all('history', with_doc=True)
                  if h['doc']['resource'].endswith(videofile)]:
            for e in [e['doc'] for e in sickrage.srCore.mainDB.db.get_many('tv_episodes', h['showid'], with_doc=True)
                      if h['season'] == e['doc']['season']
                      and h['episode'] == e['doc']['episode']
                      and e['doc']['status'] in Quality.DOWNLOADED]:

                # If we find a showid, a season number, and one or more episode numbers then we need to use those in the query
                if parse_result and (
                                parse_result.indexerid and parse_result.episode_numbers and parse_result.season_number):
                    if e['showid'] == int(parse_result.indexerid) and e['season'] == int(
                                    parse_result.season_number and e['episode']) == int(
                        parse_result.episode_numbers[0]):
                        return True
                else:
                    return True

    return False


def process_media(processPath, videoFiles, nzbName, process_method, force, is_priority, result):
    """
    Postprocess mediafiles

    :param processPath: Path to postprocess in
    :param videoFiles: Filenames to look for and postprocess
    :param nzbName: Name of NZB file related
    :param process_method: auto/manual
    :param force: Postprocess currently postprocessing file
    :param is_priority: Boolean, is this a priority download
    :param result: Previous results
    """

    processor = None
    for cur_video_file in videoFiles:
        cur_video_file_path = os.path.join(processPath, cur_video_file)

        if already_postprocessed(processPath, cur_video_file, force, result):
            result.output += logHelper("Already Processed " + cur_video_file_path + " : Skipping",
                                       sickrage.srCore.srLogger.DEBUG)
            continue

        try:
            processor = post_processor.PostProcessor(cur_video_file_path, nzbName, process_method, is_priority)
            result.result = processor.process
            process_fail_message = ""
        except EpisodePostProcessingFailedException as e:
            result.result = False
            process_fail_message = e.message

        if processor:
            result.output += processor.log

        if result.result:
            result.output += logHelper("Processing succeeded for " + cur_video_file_path)
        else:
            result.output += logHelper("Processing failed for " + cur_video_file_path + ": " + process_fail_message,
                                       sickrage.srCore.srLogger.WARNING)
            result.missedfiles.append(cur_video_file_path + " : Processing failed: " + process_fail_message)
            result.aggresult = False


def get_path_dir_files(dirName, nzbName, proc_type):
    """
    Get files in a path

    :param dirName: Directory to start in
    :param nzbName: NZB file, if present
    :param proc_type: auto/manual
    :return: a tuple of (path,dirs,files)
    """
    path = ""
    dirs = []
    files = []

    if dirName == sickrage.srCore.srConfig.TV_DOWNLOAD_DIR and not nzbName or proc_type == "manual":  # Scheduled Post Processing Active
        # Get at first all the subdir in the dirName
        for path, dirs, files in os.walk(dirName):
            break
    else:
        path, dirs = os.path.split(dirName)  # Script Post Processing
        if not (nzbName is None or nzbName.endswith('.nzb')):
            if os.path.isfile(os.path.join(dirName, nzbName)):
                dirs = []
                files = [os.path.join(dirName, nzbName)]
            else:
                dirs = [dirs]
                files = []
        else:
            dirs = [dirs]
            files = []

    return path, dirs, files


def process_failed(dirName, nzbName, result):
    """Process a download that did not complete correctly"""

    if sickrage.srCore.srConfig.USE_FAILED_DOWNLOADS:
        processor = None

        try:
            processor = failed_processor.FailedProcessor(dirName, nzbName)
            result.result = processor.process()
            process_fail_message = ""
        except FailedPostProcessingFailedException as e:
            result.result = False
            process_fail_message = e

        if processor:
            result.output += processor.log

        if sickrage.srCore.srConfig.DELETE_FAILED and result.result:
            if delete_folder(dirName, check_empty=False):
                result.output += logHelper("Deleted folder: " + dirName, sickrage.srCore.srLogger.DEBUG)

        if result.result:
            result.output += logHelper("Failed Download Processing succeeded: (" + str(nzbName) + ", " + dirName + ")")
        else:
            result.output += logHelper("Failed Download Processing failed: ({}, {}): {}"
                                       .format(nzbName, dirName, process_fail_message),
                                       sickrage.srCore.srLogger.WARNING)



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.


from __future__ import unicode_literals

import time
import urllib
import urllib2

import sickrage
from sickrage.core.common import NOTIFY_GIT_UPDATE, NOTIFY_GIT_UPDATE_TEXT, \
    notifyStrings, NOTIFY_SNATCH, NOTIFY_DOWNLOAD, NOTIFY_SUBTITLE_DOWNLOAD
from sickrage.notifiers import srNotifiers

API_URL = "https://boxcar.io/devices/providers/fWc4sgSmpcN6JujtBmR6/notifications"


class BoxcarNotifier(srNotifiers):
    def __init__(self):
        super(BoxcarNotifier, self).__init__()
        self.name = 'boxcar'

    def test_notify(self, boxcar_username):
        return self._notifyBoxcar("This is a test notification from Sick Beard", "Test", boxcar_username, force=True)

    def _sendBoxcar(self, msg, title, email, subscribe=False):
        """
        Sends a boxcar notification to the address provided

        msg: The message to send (unicode)
        title: The title of the message
        email: The email address to send the message to (or to subscribe with)
        subscribe: If true then instead of sending a message this function will send a subscription notification (optional, default is False)

        returns: True if the message succeeded, False otherwise
        """

        # build up the URL and parameters
        msg = msg.strip()
        curUrl = API_URL

        # if this is a subscription notification then act accordingly
        if subscribe:
            data = urllib.urlencode({'email': email})
            curUrl += "/subscribe"

        # for normal requests we need all these parameters
        else:
            data = urllib.urlencode({
                'email': email,
                'notification[from_screen_name]': title,
                'notification[message]': msg.encode('utf-8'),
                'notification[from_remote_service_id]': int(time.time())
            })

        # send the request to boxcar
        try:
            req = urllib2.Request(curUrl)
            handle = urllib2.urlopen(req, data)
            handle.close()

        except urllib2.HTTPError as e:
            # if we get an error back that doesn't have an error code then who knows what's really happening
            if not hasattr(e, 'code'):
                sickrage.srCore.srLogger.error("Boxcar notification failed. Error code: {}".format(e.message))
                return False
            else:
                sickrage.srCore.srLogger.warning("Boxcar notification failed. Error code: " + str(e.code))

            # HTTP status 404 if the provided email address isn't a Boxcar user.
            if e.code == 404:
                sickrage.srCore.srLogger.warning(
                    "Username is wrong/not a boxcar email. Boxcar will send an email to it")
                return False

            # For HTTP status code 401's, it is because you are passing in either an invalid token, or the user has not added your service.
            elif e.code == 401:

                # If the user has already added your service, we'll return an HTTP status code of 401.
                if subscribe:
                    sickrage.srCore.srLogger.error("Already subscribed to service")
                    # i dont know if this is true or false ... its neither but i also dont know how we got here in the first place
                    return False

                # HTTP status 401 if the user doesn't have the service added
                else:
                    subscribeNote = self._sendBoxcar(msg, title, email, True)
                    if subscribeNote:
                        sickrage.srCore.srLogger.debug("Subscription send")
                        return True
                    else:
                        sickrage.srCore.srLogger.error("Subscription could not be send")
                        return False

            # If you receive an HTTP status code of 400, it is because you failed to send the proper parameters
            elif e.code == 400:
                sickrage.srCore.srLogger.error("Wrong data sent to boxcar")
                return False

        sickrage.srCore.srLogger.info("Boxcar notification successful.")
        return True

    def _notify_snatch(self, ep_name, title=notifyStrings[NOTIFY_SNATCH]):
        if sickrage.srCore.srConfig.BOXCAR_NOTIFY_ONSNATCH:
            self._notifyBoxcar(title, ep_name)

    def _notify_download(self, ep_name, title=notifyStrings[NOTIFY_DOWNLOAD]):
        if sickrage.srCore.srConfig.BOXCAR_NOTIFY_ONDOWNLOAD:
            self._notifyBoxcar(title, ep_name)

    def _notify_subtitle_download(self, ep_name, lang, title=notifyStrings[NOTIFY_SUBTITLE_DOWNLOAD]):
        if sickrage.srCore.srConfig.BOXCAR_NOTIFY_ONSUBTITLEDOWNLOAD:
            self._notifyBoxcar(title, ep_name + ": " + lang)

    def _notify_version_update(self, new_version="??"):
        if sickrage.srCore.srConfig.USE_BOXCAR:
            update_text = notifyStrings[NOTIFY_GIT_UPDATE_TEXT]
            title = notifyStrings[NOTIFY_GIT_UPDATE]
            self._notifyBoxcar(title, update_text + new_version)

    def _notifyBoxcar(self, title, message, username=None, force=False):
        """
        Sends a boxcar notification based on the provided info or SB config

        title: The title of the notification to send
        message: The message string to send
        username: The username to send the notification to (optional, defaults to the username in the config)
        force: If True then the notification will be sent even if Boxcar is disabled in the config
        """

        if not sickrage.srCore.srConfig.USE_BOXCAR and not force:
            sickrage.srCore.srLogger.debug("Notification for Boxcar not enabled, skipping this notification")
            return False

        # if no username was given then use the one from the config
        if not username:
            username = sickrage.srCore.srConfig.BOXCAR_USERNAME

        sickrage.srCore.srLogger.debug("Sending notification for " + message)

        return self._sendBoxcar(message, title, username)



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import os

import sickrage
from sickrage.core.databases import srDatabase
from sickrage.core.databases.main.index import MainTVShowsIndex, MainTVEpisodesIndex, MainIMDBInfoIndex, \
    MainXEMRefreshIndex, MainSceneNumberingIndex, MainIndexerMappingIndex, MainHistoryIndex, MainInfoIndex, \
    MainBlacklistIndex, MainWhitelistIndex


class MainDB(srDatabase):
    _indexes = {
        'tv_shows': MainTVShowsIndex,
        'tv_episodes': MainTVEpisodesIndex,
        'imdb_info': MainIMDBInfoIndex,
        'xem_refresh': MainXEMRefreshIndex,
        'scene_numbering': MainSceneNumberingIndex,
        'indexer_mapping': MainIndexerMappingIndex,
        'info': MainInfoIndex,
        'blacklist': MainBlacklistIndex,
        'whitelist': MainWhitelistIndex,
        'history': MainHistoryIndex,
    }

    _migrate_list = {
        'tv_shows': ['indexer_id', 'indexer', 'show_name', 'location', 'network', 'genre',
                     'classification', 'runtime', 'quality', 'airs', 'status', 'flatten_folders', 'paused',
                     'startyear', 'air_by_date', 'lang', 'subtitles', 'notify_list', 'imdb_id',
                     'dvdorder', 'archive_firstmatch', 'rls_require_words', 'rls_ignore_words', 'sports', 'anime',
                     'scene', 'default_ep_status'],
        'tv_episodes': ['showid', 'indexerid', 'indexer', 'name', 'season', 'episode', 'scene_season', 'scene_episode',
                        'description', 'airdate', 'hasnfo', 'hastbn', 'status', 'location', 'file_size', 'release_name',
                        'subtitles', 'subtitles_searchcount', 'subtitles_lastsearch', 'is_proper', 'absolute_number',
                        'scene_absolute_number', 'version', 'release_group'],
        'history': ['action', 'date', 'showid', 'season', 'episode', 'quality', 'resource', 'provider', 'version'],
        'imdb_info': ['indexer_id', 'imdb_id', 'title', 'year', 'akas', 'runtimes', 'genres', 'countries',
                      'country_codes', 'certificates', 'rating', 'votes', 'last_update'],
        'info': ['last_backlog', 'last_indexer', 'last_proper_search'],
        'scene_numbering': ['indexer', 'indexer_id', 'season', 'episode', 'scene_season', 'scene_episode',
                            'absolute_number', 'scene_absolute_number'],
        'blacklist': ['show_id', 'range', 'keyword'],
        'whitelist': ['show_id', 'range', 'keyword'],
        'xem_refresh': ['indexer', 'indexer_id', 'last_refreshed'],
        'indexer_mapping': ['indexer_id', 'indexer', 'mindexer_id', 'mindexer'],
    }

    def __init__(self, name='main'):
        super(MainDB, self).__init__(name)
        self.old_db_path = os.path.join(sickrage.DATA_DIR, 'sickrage.db')

    def cleanup(self):
        self.fix_dupe_shows()
        self.fix_dupe_episodes()
        self.fix_orphaned_episodes()

    def fix_dupe_shows(self):
        checked = []

        for show in [x['doc'] for x in self.db.all('tv_shows', with_doc=True)]:
            if show['indexer_id'] in checked:
                continue

            for dupe in [x['doc'] for x in self.db.get_many('tv_shows', show['indexer_id'], with_doc=True)][1::]:
                sickrage.srCore.srLogger.info("Deleting duplicate show with id: {}".format(dupe["indexer_id"]))
                self.db.delete(dupe)

            checked += [show['indexer_id']]

    def fix_dupe_episodes(self):
        checked = []

        for ep in [x['doc'] for x in self.db.all('tv_episodes', with_doc=True)]:
            if ep['showid'] in checked:
                continue

            for dupe in [x['doc'] for x in self.db.get_many('tv_episodes', ep['showid'], with_doc=True) if
                         x['doc']['indexerid'] == ep['indexerid']][1::]:
                sickrage.srCore.srLogger.info("Deleting duplicate episode with id: {}".format(dupe["indexerid"]))
                self.db.delete(dupe)

            checked += [ep['showid']]

    def fix_orphaned_episodes(self):
        for ep in [x['doc'] for x in self.db.all('tv_episodes', with_doc=True)]:
            if not self.db.get('tv_shows', ep['showid'], with_doc=True)['doc']:
                sickrage.srCore.srLogger.info("Deleting orphan episode with id: {}".format(ep["indexerid"]))
                self.db.delete(ep)



#
# This file is part of aDBa.
#
# aDBa is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# aDBa is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with aDBa.  If not, see <http://www.gnu.org/licenses/>.
from random import shuffle


class AniDBMaper:
    blacklist = ('unused', 'retired', 'reserved')

    def getAnimeBitsA(self, amask):
        map = self.getAnimeMapA()
        return self._getBitChain(map, amask)

    def getAnimeCodesA(self, aBitChain):
        amap = self.getAnimeMapA()
        return self._getCodes(amap, aBitChain)

    def getFileBitsF(self, fmask):
        fmap = self.getFileMapF()
        return self._getBitChain(fmap, fmask)

    def getFileCodesF(self, bitChainF):
        fmap = self.getFileMapF()
        return self._getCodes(fmap, bitChainF)

    def getFileBitsA(self, amask):
        amap = self.getFileMapA()
        return self._getBitChain(amap, amask)

    def getFileCodesA(self, bitChainA):
        amap = self.getFileMapA()
        return self._getCodes(amap, bitChainA)

    def _getBitChain(self, map, wanted):
        """Return an hex string with the correct bit set corresponding to the wanted fields in the map
        """
        bit = 0
        for index, field in enumerate(map):
            if field in wanted and not field in self.blacklist:
                bit ^= 1 << len(map) - index - 1

        bit = str(bit).lstrip("0x").rstrip("L")
        bit = ''.join(["0" for unused in xrange(len(map) / 4 - len(bit))]) + bit
        return bit

    def _getCodes(self, map, bitChain):
        """Returns a list with the corresponding fields as set in the bitChain (hex string)
        """
        codeList = []
        bitChain = int(bitChain, 16)
        mapLength = len(map)
        for i in reversed(range(mapLength)):
            if bitChain & (2 ** i):
                codeList.append(map[mapLength - i - 1])
        return codeList

    def getAnimeMapA(self):
        # each line is one byte
        # only chnage this if the api changes
        map = ['aid', 'unused', 'year', 'type', 'related_aid_list', 'related_aid_type', 'category_list',
               'category_weight_list',
               'romaji_name', 'kanji_name', 'english_name', 'other_name', 'short_name_list', 'synonym_list', 'retired',
               'retired',
               'episodes', 'highest_episode_number', 'special_ep_count', 'air_date', 'end_date', 'url', 'picname',
               'category_id_list',
               'rating', 'vote_count', 'temp_rating', 'temp_vote_count', 'average_review_rating', 'review_count',
               'award_list', 'is_18_restricted',
               'anime_planet_id', 'ANN_id', 'allcinema_id', 'AnimeNfo_id', 'unused', 'unused', 'unused',
               'date_record_updated',
               'character_id_list', 'creator_id_list', 'main_creator_id_list', 'main_creator_name_list', 'unused',
               'unused', 'unused', 'unused',
               'specials_count', 'credits_count', 'other_count', 'trailer_count', 'parody_count', 'unused', 'unused',
               'unused']
        return map

    def getFileMapF(self):
        # each line is one byte
        # only chnage this if the api changes
        map = ['unused', 'aid', 'eid', 'gid', 'mylist_id', 'list_other_episodes', 'IsDeprecated', 'state',
               'size', 'ed2k', 'md5', 'sha1', 'crc32', 'unused', 'unused', 'reserved',
               'quality', 'source', 'audio_codec_list', 'audio_bitrate_list', 'video_codec', 'video_bitrate',
               'video_resolution', 'file_type_extension',
               'dub_language', 'sub_language', 'length_in_seconds', 'description', 'aired_date', 'unused', 'unused',
               'anidb_file_name',
               'mylist_state', 'mylist_filestate', 'mylist_viewed', 'mylist_viewdate', 'mylist_storage',
               'mylist_source', 'mylist_other', 'unused']
        return map

    def getFileMapA(self):
        # each line is one byte
        # only chnage this if the api changes
        map = ['anime_total_episodes', 'highest_episode_number', 'year', 'type', 'related_aid_list', 'related_aid_type',
               'category_list', 'reserved',
               'romaji_name', 'kanji_name', 'english_name', 'other_name', 'short_name_list', 'synonym_list', 'retired',
               'retired',
               'epno', 'ep_name', 'ep_romaji_name', 'ep_kanji_name', 'episode_rating', 'episode_vote_count', 'unused',
               'unused',
               'group_name', 'group_short_name', 'unused', 'unused', 'unused', 'unused', 'unused',
               'date_aid_record_updated']
        return map

    def checkMapping(self, verbos=False):

        print "------"
        print "File F: " + str(self.checkMapFileF(verbos))
        print "------"
        print "File A: " + str(self.checkMapFileA(verbos))

    def checkMapFileF(self, verbos=False):
        getGeneralMap = lambda: self.getFileMapF()
        getBits = lambda x: self.getFileBitsF(x)
        getCodes = lambda x: self.getFileCodesF(x)
        return self._checkMapGeneral(getGeneralMap, getBits, getCodes, verbos=verbos)

    def checkMapFileA(self, verbos=False):
        getGeneralMap = lambda: self.getFileMapA()
        getBits = lambda x: self.getFileBitsA(x)
        getCodes = lambda x: self.getFileCodesA(x)
        return self._checkMapGeneral(getGeneralMap, getBits, getCodes, verbos=verbos)

    def _checkMapGeneral(self, getGeneralMap, getBits, getCodes, verbos=False):
        map = getGeneralMap()
        shuffle(map)
        mask = [elem for elem in map if elem not in self.blacklist][:5]
        bits = getBits(mask)
        mask_re = getCodes(bits)
        bits_re = getBits(mask_re)
        if verbos:
            print mask
            print mask_re
            print bits
            print bits_re
            print "bits are:" + str((bits_re == bits))
            print "map is :" + str((sorted(mask_re) == sorted(mask)))
        return (bits_re == bits) and sorted(mask_re) == sorted(mask)



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import re

import sickrage
from sickrage.core.caches.tv_cache import TVCache
from sickrage.core.helpers import bs4_parser, convert_size
from sickrage.providers import TorrentProvider


class TORRENTZProvider(TorrentProvider):
    def __init__(self):

        super(TORRENTZProvider, self).__init__("Torrentz", 'http://torrentz2.eu', False)

        self.supports_backlog = True
        self.confirmed = True
        self.ratio = None
        self.minseed = None
        self.minleech = None

        self.cache = TVCache(self, min_time=15)

        self.urls.update({
            'verified': '{base_url}/feed_verified'.format(base_url=self.urls['base_url']),
            'feed': '{base_url}/feed'.format(base_url=self.urls['base_url'])
        })

    def seed_ratio(self):
        return self.ratio

    @staticmethod
    def _split_description(description):
        match = re.findall(r'[0-9]+', description)
        return int(match[0]) * 1024 ** 2, int(match[1]), int(match[2])

    def search(self, search_strings, search_mode='eponly', epcount=0, age=0, epObj=None):
        results = []

        for mode in search_strings:
            items = []
            sickrage.srCore.srLogger.debug('Search Mode: {}'.format(mode))
            for search_string in search_strings[mode]:
                search_url = self.urls['feed']
                if mode != 'RSS':
                    sickrage.srCore.srLogger.debug('Search string: {}'.format(search_string))

                try:
                    data = sickrage.srCore.srWebSession.get(search_url, params={'f': search_string}, cache=False).text
                except Exception:
                    sickrage.srCore.srLogger.debug('No data returned from provider')
                    continue

                if not data.startswith('<?xml'):
                    sickrage.srCore.srLogger.info('Expected xml but got something else, is your mirror failing?')
                    continue

                with bs4_parser(data) as parser:
                    for item in parser('item'):
                        if item.category and 'tv' not in item.category.get_text(strip=True):
                            continue

                        title = item.title.get_text(strip=True)
                        t_hash = item.guid.get_text(strip=True).rsplit('/', 1)[-1]

                        if not all([title, t_hash]):
                            continue

                        download_url = "magnet:?xt=urn:btih:" + t_hash + "&dn=" + title
                        torrent_size, seeders, leechers = self._split_description(item.find('description').text)
                        size = convert_size(torrent_size) or -1

                        # Filter unseeded torrent
                        if seeders < self.minseed or leechers < self.minleech:
                            if mode != 'RSS':
                                sickrage.srCore.srLogger.debug(
                                    "Discarding torrent because it doesn't meet the minimum seeders or leechers: {} (S:{} L:{})".format(
                                        title, seeders, leechers))
                            continue

                        items += [{
                            'title': title,
                            'link': download_url,
                            'size': size,
                            'seeders': seeders,
                            'leechers': leechers,
                            'hash': t_hash
                        }]

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: int(d.get('seeders', 0)), reverse=True)
            results += items

        return results



# Author: echel0n <echel0n@sickrage.ca>
# URL: https://sickrage.ca
# Git: https://git.sickrage.ca/SiCKRAGE/sickrage.git
#
# This file is part of SickRage.
#
# SickRage is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# SickRage is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SickRage.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import functools
import locale
from datetime import datetime

import sickrage
from sickrage.core.updaters.tz_updater import sr_timezone

date_presets = (
    '%Y-%m-%d',
    '%a, %Y-%m-%d',
    '%A, %Y-%m-%d',
    '%y-%m-%d',
    '%a, %y-%m-%d',
    '%A, %y-%m-%d',
    '%m/%d/%Y',
    '%a, %m/%d/%Y',
    '%A, %m/%d/%Y',
    '%m/%d/%y',
    '%a, %m/%d/%y',
    '%A, %m/%d/%y',
    '%m-%d-%Y',
    '%a, %m-%d-%Y',
    '%A, %m-%d-%Y',
    '%m-%d-%y',
    '%a, %m-%d-%y',
    '%A, %m-%d-%y',
    '%m.%d.%Y',
    '%a, %m.%d.%Y',
    '%A, %m.%d.%Y',
    '%m.%d.%y',
    '%a, %m.%d.%y',
    '%A, %m.%d.%y',
    '%d-%m-%Y',
    '%a, %d-%m-%Y',
    '%A, %d-%m-%Y',
    '%d-%m-%y',
    '%a, %d-%m-%y',
    '%A, %d-%m-%y',
    '%d/%m/%Y',
    '%a, %d/%m/%Y',
    '%A, %d/%m/%Y',
    '%d/%m/%y',
    '%a, %d/%m/%y',
    '%A, %d/%m/%y',
    '%d.%m.%Y',
    '%a, %d.%m.%Y',
    '%A, %d.%m.%Y',
    '%d.%m.%y',
    '%a, %d.%m.%y',
    '%A, %d.%m.%y',
    '%d. %b %Y',
    '%a, %d. %b %Y',
    '%A, %d. %b %Y',
    '%d. %b %y',
    '%a, %d. %b %y',
    '%A, %d. %b %y',
    '%d. %B %Y',
    '%a, %d. %B %Y',
    '%A, %d. %B %Y',
    '%d. %B %y',
    '%a, %d. %B %y',
    '%A, %d. %B %y',
    '%b %d, %Y',
    '%a, %b %d, %Y',
    '%A, %b %d, %Y',
    '%B %d, %Y',
    '%a, %B %d, %Y',
    '%A, %B %d, %Y'
)

time_presets = ('%I:%M:%S %p', '%H:%M:%S')


# helper class
class static_or_instance(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, instance, owner):
        return functools.partial(self.func, instance)


# subclass datetime to add function to display custom date and time formats
class srDateTime(datetime):
    has_locale = True
    en_US_norm = locale.normalize('en_US.utf-8')

    @static_or_instance
    def convert_to_setting(self, dt=None):
        try:
            if sickrage.srCore.srConfig.TIMEZONE_DISPLAY == 'local':
                return dt.astimezone(sr_timezone) if self is None else self.astimezone(sr_timezone)
            else:
                return dt if self is None else self
        except Exception:
            return dt if self is None else self

    # display Time in SickRage Format
    @static_or_instance
    def srftime(self, dt=None, show_seconds=False, t_preset=None):
        """
        Display time in SR format
        TODO: Rename this to srftime

        :param dt: datetime object
        :param show_seconds: Boolean, show seconds
        :param t_preset: Preset time format
        :return: time string
        """

        try:
            locale.setlocale(locale.LC_TIME, '')
        except Exception:
            pass

        try:
            if srDateTime.has_locale:
                locale.setlocale(locale.LC_TIME, 'en_US')
        except Exception:
            try:
                if srDateTime.has_locale:
                    locale.setlocale(locale.LC_TIME, srDateTime.en_US_norm)
            except Exception:
                srDateTime.has_locale = False

        strt = ''
        try:
            if self is None:
                if dt is not None:
                    if t_preset is not None:
                        strt = dt.strftime(t_preset)
                    elif show_seconds:
                        strt = dt.strftime(sickrage.srCore.srConfig.TIME_PRESET_W_SECONDS)
                    else:
                        strt = dt.strftime(sickrage.srCore.srConfig.TIME_PRESET)
            else:
                if t_preset is not None:
                    strt = self.strftime(t_preset)
                elif show_seconds:
                    strt = self.strftime(sickrage.srCore.srConfig.TIME_PRESET_W_SECONDS)
                else:
                    strt = self.strftime(sickrage.srCore.srConfig.TIME_PRESET)
        finally:
            try:
                if srDateTime.has_locale:
                    locale.setlocale(locale.LC_TIME, '')
            except Exception:
                srDateTime.has_locale = False

        return strt

    # display Date in SickRage Format
    @static_or_instance
    def srfdate(self, dt=None, d_preset=None):
        """
        Display date in SR format
        TODO: Rename this to srfdate

        :param dt: datetime object
        :param d_preset: Preset date format
        :return: date string
        """

        try:
            locale.setlocale(locale.LC_TIME, '')
        except Exception:
            pass

        strd = ''
        try:
            if self is None:
                if dt is not None:
                    if d_preset is not None:
                        strd = dt.strftime(d_preset)
                    else:
                        strd = dt.strftime(sickrage.srCore.srConfig.DATE_PRESET)
            else:
                if d_preset is not None:
                    strd = self.strftime(d_preset)
                else:
                    strd = self.strftime(sickrage.srCore.srConfig.DATE_PRESET)
        finally:

            try:
                locale.setlocale(locale.LC_TIME, '')
            except Exception:
                pass

        return strd

    # display Datetime in SickRage Format
    @static_or_instance
    def srfdatetime(self, dt=None, show_seconds=False, d_preset=None, t_preset=None):
        """
        Show datetime in SR format
        TODO: Rename this to srfdatetime

        :param dt: datetime object
        :param show_seconds: Boolean, show seconds as well
        :param d_preset: Preset date format
        :param t_preset: Preset time format
        :return: datetime string
        """

        try:
            locale.setlocale(locale.LC_TIME, '')
        except Exception:
            pass

        strd = ''
        try:
            if self is None:
                if dt is not None:
                    if d_preset is not None:
                        strd = dt.strftime(d_preset)
                    else:
                        strd = dt.strftime(sickrage.srCore.srConfig.DATE_PRESET)
                    try:
                        if srDateTime.has_locale:
                            locale.setlocale(locale.LC_TIME, 'en_US')
                    except Exception:
                        try:
                            if srDateTime.has_locale:
                                locale.setlocale(locale.LC_TIME, srDateTime.en_US_norm)
                        except Exception:
                            srDateTime.has_locale = False
                    if t_preset is not None:
                        strd += ', {}'.format(dt.strftime(t_preset))
                    elif show_seconds:
                        strd += ', {}'.format(dt.strftime(sickrage.srCore.srConfig.TIME_PRESET_W_SECONDS))
                    else:
                        strd += ', {}'.format(dt.strftime(sickrage.srCore.srConfig.TIME_PRESET))
            else:
                if d_preset is not None:
                    strd = self.strftime(d_preset)
                else:
                    strd = self.strftime(sickrage.srCore.srConfig.DATE_PRESET)

                try:
                    if srDateTime.has_locale:
                        locale.setlocale(locale.LC_TIME, 'en_US')
                except Exception:
                    try:
                        if srDateTime.has_locale:
                            locale.setlocale(locale.LC_TIME, srDateTime.en_US_norm)
                    except Exception:
                        srDateTime.has_locale = False
                if t_preset is not None:
                    strd += ', {}'.format(dt.strftime(t_preset))
                elif show_seconds:
                    strd += ', {}'.format(dt.strftime(sickrage.srCore.srConfig.TIME_PRESET_W_SECONDS))
                else:
                    strd += ', {}'.format(dt.strftime(sickrage.srCore.srConfig.TIME_PRESET))
        finally:
            try:
                if srDateTime.has_locale:
                    locale.setlocale(locale.LC_TIME, '')
            except Exception:
                srDateTime.has_locale = False

        return strd
